import argparse
import bs4
import json
import io
import os
import requests
import zipfile

class Scraper():
    """A scraper with which to scrape Scratch projects.

    Typical usage example:

      from ccl_scratch_tools import Scraper
        
      scraper = Scraper()
      
      project = scraper.download_project(555555555)
    """

    def __init__(self, studio_url = None, project_url = None, project_meta_url = None, comments_url = None, user_url = None, studio_meta_url = None):
        """Initializes scraper with studio and project URLs."""
        if studio_url is None:
            self.STUDIO_URL = "https://api.scratch.mit.edu/studios/{0}/projects?limit=40&offset={1}"
        else:
            self.STUDIO_URL = studio_url
        
        if studio_meta_url is None:
            self.STUDIO_META_URL = "https://api.scratch.mit.edu/studios/{0}"
        else:
            self.STUDIO_META_URL = studio_meta_url

        if project_url is None:
            self.PROJECT_URL = "https://projects.scratch.mit.edu/{0}"
        else:
            self.PROJECT_URL = project_url

        if project_meta_url is None:
            self.PROJECT_META_URL = "https://api.scratch.mit.edu/projects/{0}"
        else:
            self.PROJECT_META_URL = project_meta_url

        if comments_url is None:
            self.COMMENTS_URL = "https://scratch.mit.edu/site-api/comments/project/{0}/?page={1}"
        else:
            self.COMMENTS_URL = comments_url

        if user_url is None:
            self.USER_URL = "https://api.scratch.mit.edu/users/{0}"
        else:
            self.USER_URL = user_url

    def download_project(self, id):
        """Downloads an individual project JSON and returns it as a Python object.
        
        Args:
            id: An integer Scratch project ID.

        Returns:
            A dictionary object representing the Scratch project JSON.

        Raises:
            RuntimeError: An error occurred accessing the Scratch API, or
                the project couldn't be downloaded in/converted to JSON format.
        """
        url = self.PROJECT_URL.format(id)
        r = requests.get(url)

        if r.status_code != 200:
            raise RuntimeError("GET {0} failed with status code {1}".format(url, r.status_code))

        project = ""
        try:
            project = r.json()
        except:
            # In some cases, a binary archive will download -- handle that
            if json.decoder.JSONDecodeError:
                try:
                    f = io.BytesIO(r.content)
                    archive = zipfile.ZipFile(f)
                    if "project.json" in archive.namelist():
                        proj = archive.read("project.json")
                        project = json.loads(proj.decode("utf-8"))
                except:
                    raise RuntimeError("Cannot handle format of project {0}".format(id))
        return project

    def download_projects(self, ids, projects_to_studio=dict(), output_directory=None, file_name=None):
        """Given project IDs, download the JSON files.
        
        Args:
            ids: array-like collection of Scratch project IDs.
            projects_to_studio: dictionary mapping project IDs to studio IDs.
                If set, creates subdirectories for each studio.
            output_directory (str): directory for output; if not set,
                defaults to current working directory.
            file_name (str): if set, combines projects into one JSON file with file_name;
                else creates a separate JSON file for each project.

        Returns:
            None.
        """
        if output_directory is None:
            output_directory = os.getcwd()

        self.make_dir(output_directory)
        projects = list()
        for id in ids:
            project = self.download_project(id)

            if len(project) < 1:
                break
            
            if file_name is None:
                od = output_directory
                if len(projects_to_studio) > 0:
                    od = "{0}/{1}".format(od, projects_to_studio[id])
                    self.make_dir(od)

                with open("{0}/{1}.json".format(od, id), "w") as f:
                    json.dump(project, f)
            else:
                projects.append(project)

        # If projects has at least one item, we should write to a single file
        if len(projects) > 0 and file_name is not None:
            with open("{0}/{1}".format(output_directory, file_name), "w") as f:
                json.dump(projects, f)

    def get_id(self, url):
        """Returns the integer ID from a string that may be a URL or an ID.
        
        Args:
            url: The string representing the URL, or ID, to be extracted.

        Returns:
            An integer ID of a Scratch object, whether a studio or project.

            In case of error, returns None.
        """
        url = url.rstrip()
        a = url.rstrip("/")
        try:
            return int(a.split("/")[-1])
        except:
            return None

    def get_ids_from_file(self, filename):
        """Returns a list of IDs from a newline-separated file. Project/studio link agnostic. Works with links and IDs.
        
        Args:
            filename: String file name of a text file with line-separated URLs or IDs.

        Returns:
            A list of integer IDs. Empty if error reading file.
        """
        ids = list()
        try:
            ids = list()
            with open(filename) as f:
                for l in f.readlines():
                    ids.append(self.get_id(l))
        except:
            pass
        return ids

    def get_project_comments(self, id):
        """Returns the comments on a given Scratch project.
        
        Args:
            id (int): a Scratch project ID.

        Returns:
            A list of dictionaries, each with keys for author, comment, and timestamp.

        Raises:
            RuntimeError: An error occurred accessing the Scratch API, or the project doesn't exist.
        """

        # This is all a remastered version of GSE-CCL/scratch-comments
        comments = list()
        page = 1
        while True:
            # Call API
            url = self.COMMENTS_URL.format(id, page)
            r = requests.get(url)
            if r.status_code == 404 and page > 1:
                break
            elif r.status_code != 200:
                raise RuntimeError("GET {0} failed with status code {1}".format(url, r.status_code))

            # Use Beautiful Soup to scrape the webpage for comments
            soup = bs4.BeautifulSoup(r.content, "html.parser")
            all_comments = soup.select(".comment")
            if len(all_comments) < 1:
                break

            # Go through each comment and clean
            for comment in all_comments:
                content = comment.select_one(".content").get_text().strip()
                if content != "[deleted]":
                    cid = int(comment["data-comment-id"])
                    user = comment.select_one(".name").get_text().strip()
                    time = comment.select_one(".time")["title"]
                    comments.append({"id": cid, "username": user, "comment": content, "timestamp": time})
            page += 1

        return comments

    def get_project_meta(self, id):
        """Returns the publicly-available metadata about a given Scratch project.
        
        Args:
            id (int): a Scratch project ID.

        Returns:
            A dictionary with the entire API response from project meta API endpoint.
                None if the studio doesn't exist.

        Raises:
            RuntimeError: An error occurred accessing the Scratch API.
        """

        url = self.PROJECT_META_URL.format(id)
        r = requests.get(url)

        if r.status_code != 200 and r.status_code != 404:
            raise RuntimeError("GET {0} failed with status code {1}".format(url, r.status_code))

        project = r.json()
        if "code" in project and project["code"] == "NotFound":
            return None
        
        return project

    def get_projects_in_studio(self, id):
        """Returns the set of project IDs contained in a given Scratch studio.
        
        Args:
            id: An integer Scratch studio ID.
            
        Returns:
            A set of project IDs.

        Raises:
            RuntimeError: An error occurred accessing the Scratch API.
        """
        offset = 0
        project_ids = set()
        while True:
            url = self.STUDIO_URL.format(id, offset)
            r = requests.get(url)

            if r.status_code != 200:
                raise RuntimeError("GET {0} failed with status code {1}".format(url, r.status_code))

            # No more projects
            projects = r.json()
            if len(projects) < 1:
                break
            else:
                for project in projects:
                    project_ids.add(project["id"])
                offset += 40
            
        return project_ids

    def get_studio_meta(self, id):
        """Returns the metadata for a given Scratch studio.
        
        Args:
            id: An integer Scratch studio ID.
            
        Returns:
            A dictionary with the studio's metadata. None if the studio doesn't exist.

        Raises:
            RuntimeError: An error occurred accessing the Scratch API.
        """

        url = self.STUDIO_META_URL.format(id)
        r = requests.get(url)

        if r.status_code != 200 and r.status_code != 404:
            raise RuntimeError("GET {0} failed with status code {1}".format(url, r.status_code))

        studio_meta = r.json()
        if "code" in studio_meta and studio_meta["code"] == "NotFound":
            return None
        
        return studio_meta

    def get_user_info(self, username):
        """Gets a Scratch user's publicly-available information.
        
        Args:
            username (str): the username to look up.

        Returns:
            A dictionary with the results of the API call.

        Raises:
            RuntimeError: An error occurred accessing the Scratch API, or the user doesn't exist.
        """

        url = self.USER_URL.format(username)
        r = requests.get(url)

        if r.status_code != 200:
            raise RuntimeError("GET {0} failed with status code {1}".format(url, r.status_code))

        return r.json()
    
    def make_dir(self, path):
        """Creates a directory given path.
        
        Args:
            path (str): A file path on the current system.

        Returns:
            True, if directory was successfully created or already existed.

        Raises:
            RuntimeError: Failed to create the directory.
        """
        try:
            os.mkdir(path)
        except OSError:
            if FileExistsError:
                return True
            else:
                raise RuntimeError("Creation of directory '{0}' failed".format(path))
        else:
            return True



from __future__ import absolute_import
from ccl_scratch_tools.scratch_to_blocks import blocks

class Visualizer():
    """A way to visualize blocks by exporting to Scratchblocks syntax.

    Typical usage example:

      from ccl_scratch_tools import Visualizer

      visualizer = Visualizer()

      scratchblocks = visualizer.generate_scratchblocks(scratch_data)
    """

    def __init__(self):
        """Initializes visualizer."""
        return

    def generate_scratchblocks(self, scratch_data):
        """Generates all blocks in a project.
    
        Args:
            scratch_data (dict): the Scratch project to process.
        Returns:
            A list of scripts in the project that can be turned into text. False if unsuccessful.
        """

        if "targets" not in scratch_data:
            return False

        return blocks.generate_scratchblocks(scratch_data)

    def generate_script(self, block_id, block_list, block_ids=None, text=False):
        """Generates a script dictionary, nesting the blocks as appropriate.
        
        Args:
            block_id (str): generate the script starting from this block.
            block_list (array-like): the full list of blocks within this target.
            block_ids (array-like) (optional): the set of block IDs that are allowed to be added.
                If not provided, then all blocks are allowed.
            text (boolean) (optonal): whether to return this in Scratchblocks text syntax.
                If not provided, defaults to False - it'll just return the script dictionary.
        
        Returns:
            A dictionary with the nesting of this script as appropriate. False if unsuccessful.
        """

        script = blocks.generate_script(block_id, block_list, block_ids=None)
        if script is None or script == False:
            return False
        
        if text:
            return blocks.block_string([script])
        else:
            return script



from __future__ import absolute_import
from . import blocks, sb3_schema
import json
from jsonschema import Draft7Validator

class Parser():
    """A parser with which to parse Scratch projects.

    Typical usage example:

      from ccl_scratch_tools import Parser

      parser = Parser()

      blocks = parser.blockify("555555555.json")
    """

    def __init__(self):
        """Initializes parser."""

        self.block_data = blocks.blocks
        self.block_ignore = blocks.ignore
        self.event_listeners = blocks.event_listeners
        self.scratch_image_source = "https://assets.scratch.mit.edu/internalapi/asset/{0}/get/"
        self.sb3_schema = sb3_schema.sb3_schema

    def blockify(self, file_name=None, scratch_data=None):
        """Gets the statistics about a Scratch project.
        
        Args:
            file_name (str): the name of the Scratch JSON file to report on. (optional)
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON. (optional)

        Returns:
            A dictionary mapping function names (as in this class) to their results.

            If the data are invalid Scratch 3, returns False.
        
        Raises:
            ValueError: If neither file_name or scratch_data parameter is set.
        """

        if file_name is None and scratch_data is None:
            raise ValueError("Either file_name or scratch_data parameter is required.")

        try:
            if file_name is not None:
                with open(file_name) as f:
                    scratch_data = json.load(f)

            if scratch_data is not None and not self.is_scratch3(scratch_data):
                return False

            results = {
                "block_comments": self.get_block_comments(scratch_data),
                "block_text": self.get_block_text(scratch_data),
                "blocks": self.get_blocks(scratch_data),
                "categories": self.get_categories(scratch_data),
                "comments": self.get_comments(scratch_data),
                "costumes": self.get_costumes(scratch_data),
                "sounds": self.get_sounds(scratch_data),
                "sprites": self.get_sprite_names(scratch_data),
                "variables": self.get_variables(scratch_data)
            }
            return results
        except:
            return False

    def get_block(self, block_id, scratch_data):
        """Returns the block object in the Scratch object given block ID.
        
        Args:
            block_id (str): the Scratch block ID in the project data structure.
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A dictionary containing the block's information from the project data structure.
            Returns False if doesn't exist or if trouble accessing the data.
        """

        if self.is_scratch3(scratch_data):
            for target in scratch_data["targets"]:
                if block_id in target["blocks"]:
                    return target["blocks"][block_id]
            return False
        else:
            return False

    def get_block_comments(self, scratch_data):
        """Gets the comments left in a Scratch project, organized by block.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A dictionary mapping blocks to a list of comments associated with them.
            If the input data is invalid, returns False.
        """

        try:
            comments = dict()
            for target in scratch_data["targets"]:
                if len(target["comments"]) > 0:
                    # Loop through blocks to see which comments go where
                    for block in target["blocks"]:
                        block = target["blocks"][block]
                        if "comment" in block and len(block["comment"]) > 0:
                            if block["opcode"] not in comments:
                                comments[block["opcode"]] = list()

                            comments[block["opcode"]].append(target["comments"][block["comment"]]["text"])
            return comments
        except:
            return False

    def get_block_name(self, opcode):
        """Gets the human-readable name of a Scratch block.
        
        Args:
            opcode (str): the Scratch opcode of the block.

        Returns:
            A string containing the block's human-readable name.

            If the opcode isn't listed in our block information, returns None.
        """

        for category in self.block_data:
            if opcode in self.block_data[category]:
                return self.block_data[category][opcode]

    def get_block_names(self, items, scratch_data=None):
        """Gets the human-readable name of a list of Scratch blocks.
        
        Args:
            items: a list of blocks to translate, whether a list of opcodes
                or a list of block IDs.
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.
                Include this only if items is a list of block IDs.

        Returns:
            A list containing the blocks' human-readable names.

            If the opcode isn't listed in our block information, that item of the list will be None.
        """

        names = list()
        if scratch_data is None:
            for item in items:
                names.append(self.get_block_name(item))
        else:
            for item in items:
                block = self.get_block(item, scratch_data)
                if block is False:
                    names.append(None)
                else:
                    names.append(self.get_block_name(block["opcode"]))

        return names

    def get_block_text(self, scratch_data):
        """Gets the user-added block text, e.g. in Say blocks.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.
                Include this only if items is a list of block IDs.
        Returns:
            A list containing the user-added block text, one element per text item.
            
            False if unsuccessful.
        """

        # Get the blocks ordered by opcode
        blocks = self.get_blocks(scratch_data)

        if not blocks:
            return False

        # The opcodes that support this type of text
        opcodes = {"looks_sayforsecs": "MESSAGE",
                   "looks_say": "MESSAGE",
                   "looks_thinkforsecs": "MESSAGE",
                   "looks_think": "MESSAGE",
                   "sensing_askandwait": "QUESTION"}
        
        # Loop through possible opcodes, adding to the list of text
        texts = list()
        for opcode in opcodes:
            if opcode in blocks:
                for block in blocks[opcode]:
                    target, i = self.get_target(block, scratch_data)
                    texts.append(target["blocks"][block]["inputs"][opcodes[opcode]][1][1])

        return texts

    def get_blocks(self, scratch_data):
        """Gets the blocks used in a Scratch project.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A dictionary mapping used block opcodes to a list of block IDs
            (the IDs of the blocks of this type). Excludes unused blocks.

            If the input data is invalid, returns False.
        """

        try:
            blocks = dict()
            for target in scratch_data["targets"]:
                for block_id in target["blocks"]:
                    block = target["blocks"][block_id]
                    if block["opcode"] not in self.block_ignore:
                        if block["opcode"] not in blocks:
                            blocks[block["opcode"]] = list()
                        blocks[block["opcode"]].append(block_id)
            return blocks
        except:
            return False

    def get_categories(self, scratch_data):
        """Gets the categories of blocks used in a Scratch project.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A dictionary mapping used block categories to the number of times they've been used.

            If the input data is invalid, returns False.
        """

        try:
            categories = dict.fromkeys(self.block_data, 0)
            for target in scratch_data["targets"]:
                for block in target["blocks"]:
                    block = target["blocks"][block]
                    for category in self.block_data:
                        if block["opcode"] in self.block_data[category]:
                            categories[category] += 1
            return categories
        except:
            return False

    def get_child_blocks(self, block_id, scratch_data):
        """Gets the child blocks of a given block.
        
        Args:
            block_id (str): The ID of the block whose children we're retrieving.
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A list of block IDs, each of which corresponds to a child block of block_id.
            The first element of the list will be the block_id argument passed in.
            Returns False if data are invalid.
        """

        if self.is_scratch3(scratch_data):
            children = [block_id]
            for target in scratch_data["targets"]:
                if block_id in target["blocks"]:
                    # If this is a block that can have substacks, like loops or conditions
                    if "SUBSTACK" in target["blocks"][block_id]["inputs"]:
                        children += self.loop_through_blocks(target["blocks"][block_id]["inputs"]["SUBSTACK"][1], scratch_data)
                    # If this is a block that doesn't have substacks but functionally operates like it does
                    elif target["blocks"][block_id]["opcode"] in self.event_listeners:
                        children += self.loop_through_blocks(target["blocks"][block_id]["next"], scratch_data)
            return children
        return False

    def get_comments(self, scratch_data):
        """Gets the comments left in a Scratch project.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A list of comments in the project.

            If the input data is invalid, returns False.
        """

        try:
            comments = list()
            for target in scratch_data["targets"]:
                for comment in target["comments"]:
                    comments.append(target["comments"][comment]["text"])
            return comments
        except:
            return False

    def get_costumes(self, scratch_data):
        """Gets the costumes used in a Scratch project.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A list of the names of the costumes used.

            If the input data is invalid, returns False.
        """

        try:
            costumes = list()
            for target in scratch_data["targets"]:
                for costume in target["costumes"]:
                    costumes.append(costume["name"])
            return costumes
        except:
            return False

    def get_sounds(self, scratch_data):
        """Gets the sounds used in a Scratch project.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A list of the names of the sounds used.

            If the input data is invalid, returns False.
        """

        try:
            sounds = list()
            for target in scratch_data["targets"]:
                for sound in target["sounds"]:
                    sounds.append(sound["name"])
            return sounds
        except:
            return False

    def get_sprite(self, block_id, scratch_data):
        """Gets the sprite with which a block is associated.
        
        Args:
            block_id (str): the Scratch block ID in the project data structure.
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A dictionary containing the sprite's information from the project data structure,
            including sprite name, current costume asset ID, and current costume image URL.
            Returns False if doesn't exist or if trouble accessing the data.
        """

        result = self.get_target(block_id, scratch_data)
        if result == False:
            return False

        target, i = result
        sprite = {
            "index": i,
            "name": target["name"],
            "costume_asset": target["costumes"][target["currentCostume"]]["assetId"],
            "costume_asset_url": self.scratch_image_source
                .format(target["costumes"][target["currentCostume"]]["md5ext"])
        }
        return sprite

    def get_sprite_names(self, scratch_data):
        """Get a list of sprite names, not including the stage targets.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.
        Returns:
            A list of sprite names. Stage targets are excluded. False if unsuccessful.
        """
        if not self.is_scratch3(scratch_data):
            return False

        sprites = list()
        for target in scratch_data["targets"]:
            if not target["isStage"]:
                sprites.append(target["name"])
        return sprites

    def get_surrounding_blocks(self, block_id, scratch_data, count=5, delve=False):
        """Gets the surrounding blocks given a block ID.
        
        Args:
            block_id (str): The ID of the block of which to capture surrounding blocks.
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.
            count (int): The maximum number of blocks you want to capture, including block_id.
                Defaults to 5.
            delve (bool): Whether we should only return child blocks if this block has children,
                as with loops or event listeners.

        Returns:
            An ordered list of the IDs of the blocks surrounding block_id,
            including block_id in the middle. Returns False if the block_id doesn't exist,
            or if the data are invalid.
        """
        
        before = (count - 1) // 2
        after = count - 1 - before

        if self.is_scratch3(scratch_data):
            for target in scratch_data["targets"]:
                if block_id in target["blocks"]:
                    children = self.get_child_blocks(block_id, scratch_data)

                    # If we just want children and we have children
                    if delve and len(children) > 1:
                        return children[0:count]
                    # If we don't want just children but we do have children
                    elif not delve and len(children) > 1:
                        before_blocks = self.loop_through_blocks(block_id, scratch_data, mode="parent")
                        before_blocks = before_blocks[1:before + 1]
                        before_blocks.reverse()

                        return before_blocks + children[0:after + 1]
                    # If we don't have children
                    else:
                        before_blocks = self.loop_through_blocks(block_id, scratch_data, mode="parent")
                        after_blocks = self.loop_through_blocks(block_id, scratch_data, mode="next")
                        before_blocks = before_blocks[1:before + 1]
                        before_blocks.reverse()

                        return before_blocks + after_blocks[0:after + 1]
        return False
        
    def get_target(self, block_id, scratch_data):
        """Returns the target a block is part of.
        
        Args:
            block_id (str): find the target this block is part of.
            scratch_data (dict): the Scratch project to search through.

        Returns:
            A tuple. First, a dictionary representing the relevant target;
            second, the index of this target in the project's target list.
            
            Returns False if unsuccessful.
        """
        if not self.is_scratch3(scratch_data):
            return False
        
        for i in range(len(scratch_data["targets"])):
            target = scratch_data["targets"][i]
            if block_id in target["blocks"]:
                return target, i
                
        return False

    def get_variables(self, scratch_data):
        """Gets the variables used in a Scratch project.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            A list of the names of the variables used.

            If the input data is invalid, returns False.
        """

        try:
            variables = list()
            for target in scratch_data["targets"]:
                for variable in target["variables"]:
                    variables.append(target["variables"][variable][0])
            return variables
        except:
            return False

    def is_scratch3(self, scratch_data):
        """Checks a supposed Scratch data file against the Scratch 3 schema.
        
        Args:
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.

        Returns:
            True if the data matches the Scratch 3 schema. Otherwise False.
        """

        return Draft7Validator(self.sb3_schema).is_valid(scratch_data)

    def loop_through_blocks(self, block_id, scratch_data, mode="next"):
        """Loops through blocks in forward or backward direction.
        
        Args:
            block_id (str): The ID of the block where we start our loop.
            scratch_data (dict): a Python dictionary representing the imported Scratch JSON.
            mode (str): The mode in which we're looping -- "next" or "parent". Defaults to "next".

        Return:
            A list of block IDs, each of which corresponds to a block related to block_id.
            The first element of the list will be the block_id argument passed in.
            Returns False if data or arguments are invalid.
        """

        if mode not in ["next", "parent"]:
            return False

        if self.is_scratch3(scratch_data):
            blocks = [block_id]
            for target in scratch_data["targets"]:
                if block_id in target["blocks"]:
                    next_block = target["blocks"][block_id][mode]
                    while next_block is not None:
                        blocks.append(next_block)
                        next_block = target["blocks"][next_block][mode]
            return blocks
            


import json
import re
import requests
import sys
import webbrowser

from urllib.parse import quote

OPEN_IN_BROWSER = True


def custom_block(block):
    """Formatting for a custom My Block"""
    proccode = block["mutation"]["proccode"].replace("%s", "{}").replace("%b", "{}")
    placeholders = re.findall("%[sb]", block["mutation"]["proccode"])
    inputs = json.loads(block["mutation"]["argumentids"])
    for i, input_id in enumerate(inputs):
        if input_id not in block["inputs"]:
            block["inputs"][input_id] = [1, [10, ""]] if placeholders[i] == "%s" else [1, ["BOOL", ""]]
    return proccode, inputs


FIELDS = {
    "_mouse_": "mouse-pointer",
    "_random_": "random position",
    "PAN": "pan left/right",
}

BLOCKS = {

    # Motion
    "motion_movesteps": ("move {} steps", ["STEPS"]),
    "motion_turnright": ("turn right {} degrees", ["DEGREES"]),
    "motion_turnleft": ("turn left {} degrees", ["DEGREES"]),
    "motion_goto": ("go to {}", ["TO"]),
    "motion_gotoxy": ("go to x: {} y: {}", ["X", "Y"]),
    "motion_glideto": ("glide {} secs to {}", ["SECS", "TO"]),
    "motion_glidesecstoxy": ("glide {} secs to x: {} y: {}", ["SECS", "X", "Y"]),
    "motion_pointindirection": ("point in direction {}", ["DIRECTION"]),
    "motion_pointtowards": ("point towards {}", ["TOWARDS"]),
    "motion_changexby": ("change x by {}", ["DX"]),
    "motion_setx": ("set x to {}", ["X"]),
    "motion_changeyby": ("change y by {}", ["DY"]),
    "motion_sety": ("set y to {}", ["Y"]),
    "motion_ifonedgebounce": ("if on edge, bounce", []),
    "motion_setrotationstyle": ("set rotation style [{} v]", [["STYLE", {}]]),

    # Looks
    "looks_sayforsecs": ("say {} for {} seconds", ["MESSAGE", "SECS"]),
    "looks_say": ("say {}", ["MESSAGE"]),
    "looks_thinkforsecs": ("think {} for {} seconds", ["MESSAGE", "SECS"]),
    "looks_think": ("think {}", ["MESSAGE"]),
    "looks_switchcostumeto": ("switch costume to {}", ["COSTUME"]),
    "looks_nextcostume": ("next costume", []),
    "looks_switchbackdropto": ("switch backdrop to {}", ["BACKDROP"]),
    "looks_nextbackdrop": ("next backdrop", []),
    "looks_changesizeby": ("change size by {}", ["CHANGE"]),
    "looks_setsizeto": ("set size to {} %", ["SIZE"]),
    "looks_changeeffectby": ("change [{} v] effect by {}", [["EFFECT", {}], "CHANGE"]),
    "looks_seteffectto": ("set [{} v] effect to {}", [["EFFECT", {}], "VALUE"]),
    "looks_cleargraphiceffects": ("clear graphic effects", []),
    "looks_show": ("show", []),
    "looks_hide": ("hide", []),
    "looks_gotofrontback": ("go to [{} v] layer", [["FRONT_BACK", {}]]),
    "looks_goforwardbackwardlayers": ("go [{} v] {} layers", [["FORWARD_BACKWARD", {}], "NUM"]),

    # Sound
    "sound_playuntildone": ("play sound {} until done", ["SOUND_MENU"]),
    "sound_play": ("start sound {}", ["SOUND_MENU"]),
    "sound_stopallsounds": ("stop all sounds", []),
    "sound_changeeffectby": ("change [{} v] effect by {}", [["EFFECT", FIELDS], "VALUE"]),
    "sound_seteffectto": ("set [{} v] effect to {}", [["EFFECT", FIELDS], "VALUE"]),
    "sound_cleareffects": ("clear sound effects", []),
    "sound_changevolumeby": ("change volume by {}", ["VOLUME"]),
    "sound_setvolumeto": ("set volume to {} %", ["VOLUME"]),

    # Events
    "event_whenflagclicked": ("when flag clicked", []),
    "event_whenkeypressed": ("when [{} v] key pressed", [["KEY_OPTION", {}]]),
    "event_whenthisspriteclicked": ("when this sprite clicked", []),
    "event_whenbackdropswitchesto": ("when backdrop switches to [{} v]", [["BACKDROP", {}]]),
    "event_whengreaterthan": ("when [{} v] > {}", [["WHENGREATERTHANMENU", FIELDS], "VALUE"]),
    "event_whenbroadcastreceived": ("when I receive [{} v]", [["BROADCAST_OPTION", {}]]),
    "event_whenstageclicked": ("when stage clicked", []),

    # check these two
    "event_broadcast": ("broadcast {}", ["BROADCAST_INPUT"]),
    "event_broadcastandwait": ("broadcast {} and wait", ["BROADCAST_INPUT"]),

    # Control
    "control_wait": ("wait {} seconds", ["DURATION"]),
    "control_repeat": ("repeat {}", ["TIMES"]),
    "control_if": ("if {} then", ["CONDITION"]),
    "control_if_else": ("if {} then", ["CONDITION"]),
    "control_forever": ("forever", []),
    "control_repeat_until": ("repeat until {}", ["CONDITION"]),
    "control_stop": ("stop [{} v]", [["STOP_OPTION", {}]]),
    "control_start_as_clone": ("when I start as a clone", []),
    "control_create_clone_of": ("create clone of {}", ["CLONE_OPTION"]),
    "control_delete_this_clone": ("delete this clone", []),
    "control_wait_until": ("wait until {}", ["CONDITION"]),


    # Sensing
    "sensing_askandwait": ("ask {} and wait", ["QUESTION"]),
    "sensing_setdragmode": ("set drag mode [{} v]", [["DRAG_MODE", FIELDS]]),
    "sensing_resettimer": ("reset timer", []),


    # Variables
    "data_variable": ("({})", ["VARIABLE"]),
    "data_setvariableto": ("set [{} v] to {}", [["VARIABLE", FIELDS], "VALUE"]),
    "data_changevariableby": ("change [{} v] by {}", [["VARIABLE", FIELDS], "VALUE"]),
    "data_hidevariable": ("hide variable [{} v]", [["VARIABLE", FIELDS]]),
    "data_showvariable": ("show variable [{} v]", [["VARIABLE", FIELDS]]),

    # Lists
    "data_listcontents": ("<[{} v] contains {} >", [["LIST", FIELDS], "ITEM"]),
    "data_addtolist": ("add {} to [{} v]", ["ITEM", ["LIST", FIELDS]]),
    "data_deleteoflist": ("delete {} of [{} v]", ["INDEX", ["LIST", FIELDS]]),
    "data_deletealloflist": ("delete all of [{} v]", [["LIST", FIELDS]]),
    "data_insertatlist": ("insert {} at {} of [{} v] ", ["ITEM", "INDEX", ["LIST", FIELDS]]),
    "data_replaceitemoflist": ("replace item {} of [{} v] with {}", ["INDEX", ["LIST", FIELDS], "ITEM"]),
    "data_hidelist": ("show list [{} v]", [["LIST", FIELDS]]),
    "data_showlist": ("hide list [{} v]", [["LIST", FIELDS]]),

    # My Blocks
    "procedures_definition": ("define {}", ["custom_block"]),
    "procedures_call": custom_block,

    # Pen
    "pen_clear": ("erase all", []),
    "pen_stamp": ("stamp", []),
    "pen_penDown": ("pen down", []),
    "pen_penUp": ("pen up", []),
    "pen_setPenColorToColor": ("set pen color to {}", ["COLOR"]),
    "pen_changePenColorParamBy": ("change pen ({} v) by {}", ["COLOR_PARAM", "VALUE"]),
    "pen_setPenColorParamTo": ("set pen ({} v) to {}", ["COLOR_PARAM", "VALUE"]),
    "pen_changePenSizeBy": ("change pen size by {}", ["SIZE"]),
    "pen_setPenSizeTo": ("set pen size to {}", ["SIZE"]),

    # Music
    "music_playDrumForBeats": ("play drum ({} v) for {} beats", ["DRUM", "BEATS"]),
    "music_restForBeats": ("rest for {} beats", ["BEATS"]),
    "music_playNoteForBeats": ("play note ({}) for {} beats", ["NOTE", "BEATS"]), 
    "music_setInstrument": ("set instrument to ({} v)", ["INSTRUMENT"]),
    "music_setTempo": ("set tempo to {}", ["TEMPO"]),
    "music_changeTempo": ("change tempo by {}", ["TEMPO"]),

    # Video
    "videoSensing_whenMotionGreaterThan": ("when video motion > {}", ["REFERENCE"]),
    "videoSensing_videoToggle": ("turn video ({} v)", ["VIDEO_STATE"]),
    "videoSensing_setVideoTransparency": ("set video transparency to {}", ["TRANSPARENCY"]),

    # Text to Speech
    "text2speech_speakAndWait": ("speak {}", ["WORDS"]),
    "text2speech_setVoice": ("set voice to ({} v)", ["VOICE"]),
    "text2speech_setLanguage": ("set language to ({} v)", ["LANGUAGE"]),

    # Translate
    "translate_getTranslate": ("translate {} to ({} v)", ["WORDS", "LANGUAGE"]),

}

INPUTS = {
    # Motion
    "motion_goto_menu": ("({} v)", [["TO", FIELDS]]),
    "motion_glideto_menu": ("({} v)", [["TO", FIELDS]]),
    "motion_pointtowards_menu": ("({} v)", [["TOWARDS", FIELDS]]),
    "motion_xposition": ("(x position)", []),
    "motion_yposition": ("(y position)", []),
    "motion_direction": ("(direction)", []),

    # Looks
    "looks_costume": ("({} v)", [["COSTUME", {}]]),
    "looks_backdrops": ("({} v)", [["BACKDROP", {}]]),
    "looks_costumenumbername": ("(costume [{} v])", [["NUMBER_NAME", {}]]),
    "looks_backdropnumbername": ("(backdrop [{} v])", [["NUMBER_NAME", {}]]),
    "looks_size": ("(size)", []),

    # Sound
    "sound_sounds_menu": ("({} v)", [["SOUND_MENU", {"attrs": ["preservecase"]}]]),
    "sound_volume": ("(volume)", []),

    # Control -- check this block
    "control_create_clone_of_menu": ("({} v)", [["CLONE_OPTION", FIELDS]]),

    # Sensing
    "sensing_mousedown": ("<mouse down?>", []),
    "sensing_touchingobject": ("<touching [{} v]?>", ["TOUCHINGOBJECTMENU"]),
    "sensing_touchingobjectmenu": ("{}", [["TOUCHINGOBJECTMENU", FIELDS]]),
    "sensing_touchingcolor": ("<touching color {}?>", ["COLOR"]),
    "sensing_coloristouchingcolor": ("<color {} is touching {}?>", ["COLOR1", "COLOR2"]),
    "sensing_distanceto": ("(distance to [{} v])", ["DISTANCETOMENU"]),
    "sensing_distancetomenu": ("{}", [["DISTANCETOMENU", FIELDS]]),
    "sensing_keypressed": ("<key [{} v] pressed?>", ["KEY_OPTION"]),
    "sensing_keyoptions": ("{}", [["KEY_OPTION", FIELDS]]),
    "sensing_mousex": ("(mouse x)", []),
    "sensing_mousey": ("(mouse y)", []),
    "sensing_loudness": ("(loudness)", []),
    "sensing_timer": ("(timer)", []),
    "sensing_of": ("([{} v] of {})", [["PROPERTY", FIELDS], "OBJECT"]),
    "sensing_of_object_menu": ("[{} v]", [["OBJECT", FIELDS]]),
    "sensing_current": ("(current [{} v])", [["CURRENTMENU", FIELDS]]),
    "sensing_dayssince2000": ("(days since 2000)", []),
    "sensing_username": ("(username)", []),
    "sensing_answer": ("(answer)", []),

    # Operators
    "operator_add": ("({} + {})", ["NUM1", "NUM2"]),
    "operator_subtract": ("({} - {})", ["NUM1", "NUM2"]),
    "operator_equals": ("<{} = {}>", ["OPERAND1", "OPERAND2"]),
    "operator_random": ("(pick random {} to {})", ["FROM", "TO"]),
    "operator_gt": ("<{} > {}>", ["OPERAND1", "OPERAND2"]),
    "operator_lt": ("<{} < {}>", ["OPERAND1", "OPERAND2"]),
    "operator_and": ("<{} and {}>", ["OPERAND1", "OPERAND2"]),
    "operator_round": ("(round {})", ["NUM"]),
    "operator_mathop": ("([{} v] of {} )", [["OPERATOR", FIELDS], "NUM"]),
    "operator_or": ("<{} or {}>", ["OPERAND1", "OPERAND2"]),
    "operator_not": ("<not {}>", ["OPERAND"]),
    "operator_join": ("(join {} {})", ["STRING1", "STRING2"]),
    "operator_letter_of": ("(letter {} of {}", ["LETTER", "STRING"]),
    "operator_length": ("(length of {})", ["STRING"]),
    "operator_contains": ("< {} contains {}?>", ["STRING1", "STRING2"]),
    "operator_mod": ("({} mod {})", ["NUM1", "NUM2"]),
    "operator_multiply": ("({} * {})", ["NUM1", "NUM2"]),
    "operator_divide": ("({} / {})", ["NUM1", "NUM2"]),

    # List
    "data_itemoflist": ("(item {} of [{} v])", ["INDEX", ["LIST", FIELDS]]),
    "data_itemnumoflist": ("(item # of {} in [{} v])", ["ITEM", ["LIST", FIELDS]]),
    "data_lengthoflist": ("(length of [{} v])", [["LIST", FIELDS]]),
    "data_listcontainsitem": ("<[{} v] contains {}?>", [["LIST", FIELDS], "ITEM"]),

    # My Blocks
    "procedures_prototype": custom_block,
    "argument_reporter_boolean": ("<{}>", [["VALUE", {}]]),
    "argument_reporter_string_number": ("({})", [["VALUE", {}]]),

    # Pen
    "pen_menu_colorParam": ("{}", [["colorParam", FIELDS]]),

    # Music
    "music_menu_DRUM": ("{}", [["DRUM", FIELDS]]),
    "note": ("{}", [["NOTE", FIELDS]]),
    "music_menu_INSTRUMENT": ("{}", [["INSTRUMENT", FIELDS]]),
    "music_getTempo": ("(tempo)", []),

    # Video
    "videoSensing_menu_VIDEO_STATE": ("{}", [["VIDEO_STATE", FIELDS]]),
    "videoSensing_videoOn": ("(video ({} v) on ({} v))", ["ATTRIBUTE", "SUBJECT"]),
    "videoSensing_menu_ATTRIBUTE": ("{}", [["ATTRIBUTE", FIELDS]]),
    "videoSensing_menu_SUBJECT": ("{}", [["SUBJECT", FIELDS]]),

    # Text to Speech
    "text2speech_menu_voices": ("{}", [["voices", FIELDS]]),
    "text2speech_menu_languages": ("{}", [["languages", FIELDS]]),

    # Translate
    "translate_menu_languages": ("{}", [["languages", FIELDS]]),
    "translate_getViewerLanguage": ("language", []),
}


def get_project_from_url(url):
    """Return project data for a URL."""

    # Validate project URL
    match = re.search("scratch.mit.edu/projects/(\\d+)/", url)
    if match is None:
        return None
    project_id = match.group(1)

    # Query for project JSON
    res = requests.get(f"https://projects.scratch.mit.edu/{project_id}")
    return res.json()


def generate_scratchblocks(project):
    """Generates all blocks in a project.
    
    Args:
        project (dict): the Scratch project to process.
        surrounding (list): list of blocks allowed to added
    Returns:
        A list of scripts in the project that can be turned into text.
    """
    targets = project["targets"]
    scripts = []

    # Check each target for a new script
    for target in targets:
        for block_id, block in target["blocks"].items():
            is_start = (
                isinstance(block, dict) and block["parent"] is None
                and block["opcode"] in BLOCKS
            )
            if is_start:
                script = generate_script(block_id, target["blocks"])
                scripts.append(script)
    return scripts


def generate_script(block_id, blocks, block_ids=None):
    """Generates a script.
    
    Args:
        block_id (str): the start block.
        blocks (array-like): the list of blocks within this target.
        block_ids (array-like) (optional): the set of block IDs that are allowed to be added.
            If not set, then all blocks are allowed.
    
    Returns:
        A dictionary with the nesting of this script as appropriate.
    """

    # If we just want everything
    if block_ids is None:
        block_ids = blocks

    # If the current block isn't allowed, we're not adding it.
    if block_id not in block_ids:
        return None

    block = blocks[block_id]
    opcode = block["opcode"]

    # Format current block
    if opcode in BLOCKS:
        block_structure = BLOCKS[opcode]
        if callable(block_structure):
            name, inputs = block_structure(block)
        else:
            name, inputs = block_structure
        label = format_block(block_id, blocks, name, inputs)
        script = {
            "label": label
        }
    else:
        raise Exception(f"MISSING handler for {opcode}")

    # Handle next block
    next_block = block["next"]
    if next_block in block_ids:
        script["next"] = generate_script(next_block, blocks, block_ids)

    # Handle substacks
    if "SUBSTACK" in block["inputs"]:
        substack_id = block["inputs"]["SUBSTACK"][1]
        script["substack"] = generate_script(substack_id, blocks, block_ids)

    if "SUBSTACK2" in block["inputs"]:
        substack_id = block["inputs"]["SUBSTACK2"][1]
        script["substack2"] = generate_script(substack_id, blocks, block_ids)

    return script


def generate_input(input_block, blocks):
    """Generate input based on the value of INPUTS for a block."""
    main_input = input_block[1]

    # Handle inputs that are references to other blocks
    if isinstance(main_input, str):
        return generate_input_block(main_input, blocks)

    # Handle inputs that do not refer to other blocks
    elif isinstance(main_input, list):
        input_type = main_input[0]
        input_value = main_input[1]
        if input_type in [4, 5, 6, 7, 8, 9, 12, 13]:  # number, variable
            return f"({input_value})"
        elif input_type in [10, 11]:  # string, broadcast
            return f"[{input_value}]"
        elif input_type == "BOOL":
            return f"<{input_value}>"

    else:
        raise Exception(f"Missing handler for input type {type(main_input)}")


def generate_input_block(block_id, blocks):
    """Generate the value of an input reporter."""
    block = blocks[block_id]
    opcode = block["opcode"]
    if opcode in INPUTS:
        block_structure = INPUTS[opcode]
        if callable(block_structure):
            name, inputs = block_structure(block)
        else:
            name, inputs = block_structure
        input_block = format_block(block_id, blocks, name, inputs)
        return input_block
    else:
        raise Exception(f"Missing handler for input {opcode}")


def format_block(block_id, blocks, name, inputs):
    block = blocks[block_id]

    # Generate inputs for block
    args = []
    for input_name in inputs:
        if isinstance(input_name, str):
            arg = generate_input(block["inputs"][input_name], blocks)
            args.append(arg)
        elif isinstance(input_name, list):
            field_name, mapping = input_name
            args.append(get_field_name(mapping, block, field_name))
        else:
            raise Exception(f"unsupported block type {type(input_name)}")
    return name.format(*args)


def get_field_name(mapping, block, field_name):
    value = block["fields"][field_name][0]
    if mapping.get(value):
        return mapping.get(value)
    elif "preservecase" in mapping.get("attrs", []):
        return value
    else:
        return value.lower()


def block_string(scripts):
    def indent_string(block, indent):
        # If block isn't allowed, we just ignore it.
        if block is None:
            return ""

        output = " " * indent + block["label"] + "\n"

        # Print substack
        if "substack" in block:
            output += indent_string(block["substack"], indent + 4)
            if "substack2" in block:
                output += " " * indent + "else\n"
                output += indent_string(block["substack2"], indent + 4)
            output += " " * indent + "end\n"

        # Print next block
        if "next" in block:
            output += indent_string(block["next"], indent)

        return output

    output = ""
    for script in scripts:
        output += indent_string(script, 0)
        output += "\n"
    return output


def main():
    # Get Scratch project URL
    if len(sys.argv) > 2:
        sys.exit("Usage: python blocks.py SCRATCH_PROJECT_URL")
    elif len(sys.argv) == 2:
        url = sys.argv[1]
    else:
        url = input("Scratch Project URL: ")
    # Download project data
    data = get_project_from_url(url)
    if data is None:
        sys.exit("Could not download project.")
    print("data", data)
    # Generate blocks
    blocks = generate_scratchblocks(data)
    text = block_string(blocks)
    print(text)
    if OPEN_IN_BROWSER:
        data = quote(text)
        url = f"http://scratchblocks.github.io/#?style=scratch3&script={data}"
        webbrowser.open(url)


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ccl_scratch_tools",
    version="0.1.4",
    author="Creative Computing Lab",
    author_email="jarchibald@college.harvard.edu",
    description="Tools with which to work with Scratch JSON files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GSE-CCL/scratch-tools",
    project_urls={
        "Documentation": "https://ccl-scratch-tools.readthedocs.io/en/latest/"
    },
    packages=["ccl_scratch_tools", "ccl_scratch_tools.scratch_to_blocks"],
    install_requires=[
        "bs4",
        "jsonschema",
        "requests"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)


blocks = {
    "motion": {"motion_movesteps": "Move Steps", "motion_turnright": "Turn Right", "motion_turnleft": "Turn Left", "motion_goto": "Go to", "motion_goto_menu": "Go to Menu", "motion_gotoxy": "Go to XY", "motion_glideto": "Glide To", "motion_glidesecstoxy": "Glide Secs to XY", "motion_pointindirection": "Point in Direction", "motion_pointtowards_menu": "Point Towards Menu", "motion_pointtowards": "Point Towards", "motion_changexby": "Change X By", "motion_setx": "Set X", "motion_changeyby": "Change Y By", "motion_sety": "Set Y", "motion_ifonedgebounce": "If on Edge, Bounce", "motion_setrotationstyle": "Set Rotation Style", "motion_xposition": "X Position", "motion_yposition": "Y Position", "motion_direction": "Direction"},
    "control": {"control_wait": "Wait Seconds", "control_repeat": "Repeat", "control_if": "If", "control_forever": "Forever", "control_repeat_until": "Repeat Until", "control_stop": "Stop", "control_start_as_clone": "Start as Clone", "control_create_clone_of": "Create Clone of", "control_create_clone_of_menu": "Create Clone of", "control_delete_this_clone": "Delete this Clone", "control_if_else": "If Else", "control_wait_until": "Wait Until"},
    "event": {"event_whenflagclicked": "When Flag Clicked", "event_whenkeypressed": "When Key Pressed", "event_whenthisspriteclicked": "When Sprite Clicked", "event_whenbackdropswitchesto": "When Backdrop Switches To", "event_whengreaterthan": "When Greater Than", "event_whenbroadcastreceived": "When Broadcast Received", "event_broadcast": "Broadcast", "event_broadcastandwait": "Broadcast Event and Wait"},
    "looks": {"looks_sayforsecs": "Say for Secs", "looks_say": "Say", "looks_switchcostumeto": "Switch Costume", "looks_costume": "Costume", "looks_nextcostume": "Next Costume", "looks_switchbackdropto": "Switch Backdrop To", "looks_backdrops": "Backdrops", "looks_nextbackdrop": "Next Backdrop", "looks_changesizeby": "Change Size By", "looks_setsizeto": "Set Size To", "looks_changeeffectby": "Change Effect By", "looks_seteffectto": "Set Effect To", "looks_cleargraphiceffects": "Clear Graphic Effects", "looks_show": "Show", "looks_hide": "Hide", "looks_gotofrontback": "Go to Front/Back Layer", "looks_goforwardbackwardlayers": "Go Forward/Backwards Layers", "looks_costumenumbername": "Costume Number Name", "looks_backdropnumbername": "Backdrop Number Name", "looks_size": "Size", "looks_thinkforsecs": "Think for Secs", "looks_think": "Think"},
    "operators": {"operator_add": "Add", "operator_subtract": "Subtract", "operator_random": "Random", "operator_gt": "Greater Than", "operator_lt": "Less Than", "operator_equals": "Equals", "operator_and": "And", "operator_round": "Round", "operator_mathop": "Other Math Operation", "operator_or": "Or", "operator_not": "Not", "operator_join": "Join", "operator_letter_of": "Letter of", "operator_length": "Length of", "operator_contains": "Parameter Contains", "operator_mod": "Modulo", "operator_multiply": "Multiply", "operator_divide": "Divide"},
    "sensing": {"sensing_touchingobject": "Touching Object", "sensing_touchingobjectmenu": "Touching Specific Object", "sensing_touchingcolor": "Touch Color", "sensing_coloristouchingcolor": "Color is Touching Color", "sensing_distanceto": "Distance To Object", "sensing_distancetomenu": "Distance to Specific Object", "sensing_keypressed": "Key Pressed", "sensing_keyoptions": "Key Options", "sensing_mousedown": "Mouse Down", "sensing_mousex": "Mouse X", "sensing_mousey": "Mouse Y", "sensing_setdragmode": "Set Drag Mode", "sensing_loudness": "Loudness", "sensing_timer": "Timer", "sensing_resettimer": "Reset Timer", "sensing_of": "Sensing Of", "sensing_of_object_menu": "Sensing of Object", "sensing_current": "Current Time Variable", "sensing_dayssince2000": "Days Since 2000", "sensing_username": "Username", "sensing_askandwait": "Ask and Wait", "sensing_answer": "Answer"},
    "sound": {"sound_playuntildone": "Play Until Done", "sound_stopallsounds": "Stop All Sounds", "sound_changeeffectby": "Change Effect By", "sound_seteffectto": "Set Effect To", "sound_cleareffects": "Clear Effects", "sound_changevolumeby": "Change Volume By", "sound_setvolumeto": "Set Volume To", "sound_volume": "Volume", "sound_play": "Play"},
    "procedures": {"procedures_definition": "Block Definition", "procedures_call": "Block Call"},
    "data": {"data_variable": "Variable", "data_setvariableto": "Set Variable To", "data_changevariableby": "Change Variable By", "data_hidevariable": "Hide Variable", "data_showvariable": "Show Variable", "data_listcontents": "List Contents", "data_addtolist": "Add To List", "data_deleteoflist": "Delete Of List", "data_deletealloflist": "Delete All Of List", "data_insertatlist": "Insert At List", "data_replaceitemoflist": "Replace Item Of List", "data_itemoflist": "Item Of List", "data_itemnumoflist": "Item Number Of List", "data_lengthoflist": "Length Of List", "data_listcontainsitem": "List Contains", "data_hidelist": "Hide List", "data_showlist": "Show List"}
}

# Often duplicates
ignore = {
    "sound_sounds_menu": "Sounds Menu",
    "procedures_prototype": "Block Prototype"
}

# Event listeners
event_listeners = {
    "event_whenflagclicked": "When Flag Clicked",
    "event_whenkeypressed": "When Key Pressed",
    "event_whenthisspriteclicked": "When Sprite Clicked",
    "event_whenbackdropswitchesto": "When Backdrop Switches To",
    "event_whengreaterthan": "When Greater Than",
    "event_whenbroadcastreceived": "When Broadcast Received"
}